#! /usr/bin/perl -w

# CALL: ./NewGenerateGlueGrammars.perl [mbot|xtop] rule-table path/to/grammar

use strict;

my $arguments = @ARGV;

die "ERROR: Need 3 arguments!\nCall like this:\tGenerateGlueGrammars.perl [mbot|xtop] path/to/rule-table path/to/grammars\n" if $arguments < 3;

my $switch = $ARGV[0];
my $rule_table = $ARGV[1];
my $path = $ARGV[2];


print ":::" . $switch . ":::\n";
if ($rule_table =~ /\.gz$/) {
  open (TABLE, "gunzip -c $rule_table |") || die "can't open pipe to $rule_table";
}
else {
  open (TABLE, $rule_table) || die "can't open $rule_table";
}

# FIRST: GET ALL TOP AND GLUE TAGS FROM RULE TABLE
my %top_tags_source, my %top_tags_target, my %glue_tags_source, my %glue_tags_target;
my @splitter;
my $counter = 0;
my $warn = 0;

print "Getting tags from rule table...\n";
while ( <TABLE> ) {
  $counter++;
  chomp;
  $_ =~ s/\s+/ /g;
  @splitter = split('\|\|\|', $_);

  # get labels from source side
  my $source = $splitter[1];
  $source =~ s/^\s+//;
  $source =~ s/\s+$//;
  my @s_split = split(' ', $source);

  for (my $i=0; $i<$#s_split; $i++) {
    if ( $s_split[$i] =~ m/^\[(.+?)\]\[.+\]$/ ) {
      $glue_tags_source{$1}++;
    }
  }
  if ( $s_split[-1] =~  m/^\[(.+)\]$/ ) {
    $top_tags_source{$1}++;
  }
  else {
    $warn = 1;
  }

  # get labels from target side
  my $target = $splitter[2];
  $target =~ s/^\s+//;
  $target =~ s/\s+$//;

  # if mbot get top and glue tags from discontiguous target side
  if ( $switch eq "mbot" ) {
    my @t_split = split('\|\|', $target);

    for (my $i=0; $i<=$#t_split; $i++) {
      my @c_split = split(' ', $t_split[$i]);

      for (my $j=0; $j<$#c_split; $j++) {
        if ( $c_split[$j] =~ m/^\[.+?\]\[(.+?)\]/g ) {
          $glue_tags_target{$1}++;
        }
      }
      if ( $c_split[-1] =~  m/^\[(.+)\]$/ ) {
        $top_tags_target{$1}++;
      }
      else {
        $warn = 1;
      }
    }
  }
  # else get top and glue tags from contiguous target side
  elsif ( $switch eq "xtop" ) {
    my @t_split = split(' ', $target);

    for (my $i=0; $i<$#t_split; $i++) {
      if ( $t_split[$i] =~ m/^\[.+?\]\[(.+)\]$/ ) {
        $glue_tags_target{$1}++;
      }
    }
    if ( $t_split[-1] =~  m/^\[(.+)\]$/ ) {
      $top_tags_target{$1}++;
    }
    else {
      $warn = 1;
    }
  }
  if ( $warn ) {
    print "WARNING: Empty brackets at line $counter:\n" . $_ . "\n";
    $warn = 0;
  }
}

# SECOND: GENERATE GLUE-GRAMMAR
print "Generating glue-grammar...\n";
open(GLUE,"> $path/glue-grammar") or die $!;

my $prefix = "[MBOT] ||| ";
my $suffix = " ||| 1.0 ||| ||| 0";
my $suffix1 = " ||| 1.0 ||| 0-0 ||| 0 ";
my $suffix_top = " ||| 1.0 ||| 1-1 ||| 0 ";
my $suffix_glue = " ||| 2.718 ||| 0-0 1-1 ||| 0 ";

# basic rules
print GLUE $prefix . "<s> [ROOT] ||| <s> [Q]" . $suffix . "\n";
print GLUE $prefix . "[ROOT][Q] </s> [ROOT] ||| [ROOT][Q] </s> [Q]"  . $suffix1 . "\n";

# top rules
for my $top_source ( keys %top_tags_source ) {
  for my $top_target ( keys %top_tags_target ) {
    print GLUE $prefix . "<s> [" . $top_source . "][" . $top_target . "] </s> [ROOT] ||| <s> [" . $top_source . "][" . $top_target . "] </s> [Q]" . $suffix_top . "\n";
  }
}

# glue rules
for my $glue_source ( keys %glue_tags_source ) {
  for my $glue_target ( keys %glue_tags_target ) {
    print GLUE $prefix . "[ROOT][Q] [" . $glue_source . "][" . $glue_target . "] [ROOT] ||| [ROOT][Q] [" . $glue_source . "][" . $glue_target . "] [Q]" . $suffix_glue . "\n";
  }
}

# THIRD: GENERATE GLUE-MBOT-GRAMMAR IF SWITCH mbot
my %mbot_hash;
if ( $switch eq "mbot" ) {
  print "Generating glue-mbot-grammar...\n";

  my $prefix_glue = "[MBOT] ||| [ROOT][Q] ";
  my $max_size = 0;

  if ($rule_table =~ /\.gz$/) {
    open (TABLE, "gunzip -c $rule_table |") || die "can't open pipe to $rule_table";
  }
  else {
    open (TABLE, $rule_table) || die "can't open $rule_table";
  }

  while ( <TABLE> ) {
    chomp;
    my @components = split('\|\|\|', $_);

    my $source = $components[1];
    $source =~ s/^\s+//;
    $source =~ s/\s+$//;

    if ( $source !~ m/\[[^\s]+\]\[[^\s]+\]/g ) {
      $source =~ m/(\[[^\s]+\])$/;
      my $source_tag =  $1;

      my $target = $components[2];
      $target =~ s/^\s+//;
      $target =~ s/\s+$//;

      if ( $target =~ m/\|\|/ ) {
        my $newtarget = "";
        my $target_tags = "";
        my $check = 1;
        my @t_comp = split(' \|\| ', $target);
        my $t_size = @t_comp;
        if ( $max_size < $t_size ) {
          $max_size = $t_size;
        }
        foreach ( @t_comp ) {
          if ( $_ !~ m/\[[^\s]+\]\[[^\s]+\]/g ) {
            $_ =~ m/\s(\[[^\s]+\])$/;
            $target_tags .= $1;
            $newtarget .= $source_tag . $1 . " ";
          }
          else {
            $check = 0;
          }
        }
        if ( $check == 1 ) {
          $newtarget .= "[Q] ";
          $mbot_hash{$t_size}{$prefix_glue . $source_tag . $target_tags . " [ROOT] ||| [ROOT][Q] " . $newtarget}++;
        }
      }
    }
  }

  open(MBOT, "> $path/glue-mbot-grammar") or die $!;
  print "Maximum number of mbot components is $max_size !\n";
  for ( my $i=2; $i<=$max_size; $i++ ) {
    my $suffix = "||| 2.718 ||| 0-0 ";
    for ( my $j=1; $j<=$i; $j++ ) {
      $suffix .=  "1-" . $j . " ";
    }
    $suffix .= "||| 0 ";
    if ( defined $mbot_hash{$i} ) {
      my %current = ( %{ $mbot_hash{$i} } );
      for my $key ( keys %current ) {
        print MBOT $key . $suffix . "\n";
      }
    }
  }
}

