#!/usr/bin/env python

from itertools import izip
from collections import defaultdict
import sys

def tokenize(s):
    tokens = []
    current_token = ""
    backslash = False
    for c in s:
        if not backslash and c in "() ":
            if current_token:
                tokens.append(current_token)
            tokens.append(c)
            current_token = ""
        else:
            current_token += c
            if c == "\\":
                backslash = not(backslash)
            else:
                backslash = False
    return tokens


class Alignment(list):
    def __init__(self, s=""):
        list.__init__(self)
        for pair in s.split():
            l, r = pair.split("-", 1)
            self.append((int(l), int(r)))

    def length(self):
        words = zip(*self)
        return (max(words[0]) + 1, max(words[1]) + 1)

    def compatible(self, t1, t2):
        l1, l2 = self.length()
        return (l1 <= t1.length() and l2 <= t2.length())

    def __str__(self):
        return " ".join(("%s-%s" % pair) for pair in self)

class TreeSequence(list):

    def __init__(self, term=""):
        self.roots = []
        if term:
            self.add_tree(term)

    def add_tree(self, term):
        tokens = tokenize(term)
        stack = []
        index = 0
        current_token = ""
        for token in reversed(tokens):
            if token == ")": #push a bracket onto the stack
                stack.append(")")
            elif token == " ":
                if current_token: #terminal
                    stack.append(index)
                    self.append((current_token, []))
                    index += 1
                    current_token = ""
            elif token == "(":  #pop non-brackets from the stack: children
                                #replace first bracket on the stack by index
                children = []
                while True:
                    item = stack.pop()
                    if item == ")":
                        stack.append(index)
                        break
                    else:
                        children.append(item)
                self.append((current_token, children))
                current_token = ""
                index += 1
            else:
                current_token = token
        self.roots.append(len(self) - 1)

    def parent(self, pos):
        for index, node in enumerate(self):
            if pos in node[1]:
                return index
        return None

    def __str__(self):
        return " ; ".join(self.linearize(r) for r in self.roots)

    def linearize(self, index): #standard tree representation
        if index is None:
            index = self.roots[0]
        def lin(index):
            label, children = self[index]
            if children:
                ch_s = " ".join(lin(ch) for ch in children)
                return "(%s:%s %s)" % (label, index, ch_s)
            else:
                return "%s:%s" % (label, index)
        return lin(index)

    def length(self):
        return sum(1 for node in self if not node[1])


class MBOTRule:

    @classmethod
    def from_cut_points(cls, s, t, cut_points):

        def sequence(tree, roots, leaves):
            def subtree_list(index):
                label, children = tree[index]
                l = [index]
                if index in leaves:
                    return l
                else:
                    for ch in children:
                        l.extend(subtree_list(ch))
                return l
            
            m = {}
            s = TreeSequence()
            for root in roots:
                for index in reversed(subtree_list(root)):
                    m[index] = len(s)
                    label, children = tree[index]
                    if index in leaves:
                        children = []
                    s.append((label, [m[i] for i in children]))
                s.roots.append(len(s) - 1)
            return s, m

        rule = cls()
        roots, leaves = cut_points
        lhs_root, rhs_roots = roots
        lhs_leaves = [leaf[0] for leaf in leaves]
        rhs_leaves = [l for leaf in leaves for l in leaf[1]]
 
        lhs_tree, lhs_mapping = sequence(s, [lhs_root], lhs_leaves)
        rhs_tree, rhs_mapping = sequence(t, 
            reversed(sorted(rhs_roots)), rhs_leaves)

        #remap the alignment
        lex_al = set()
        nt_al = set()

        for (si, tis) in leaves:
            for ti in tis:
                al = (lhs_mapping[si], rhs_mapping[ti])
                if t[ti][1]:
                    nt_al.add(al)
                else:
                    lex_al.add(al)

        rule.lhs = lhs_tree
        rule.rhs = rhs_tree
        rule.lex_al = lex_al
        rule.nt_al = nt_al
        return rule

    def __str__(self):
        def format_tree(t, i, rhs=False):
            label, children = t[i]
            if children:
                return "(%s %s)" % (label,
                    " ".join(format_tree(t, ch, rhs) for ch in children))
            else: # leaf nonterminal or terminal
                if rhs:
                    nt = set(a[0] for a in self.nt_al if a[1] == i)
                    lex = set(a[0] for a in self.lex_al if a[1] == i)
                    nt_string = "".join("@%s" % (n,) for n in nt)
                    lex_string = "".join("@%s" % (l,) for l in lex)
                    if nt:
                        return "(%s:%s%s)" % (label, i, nt_string)
                    else:
                        return "%s:%s%s" % (label, i, lex_string)

                else:
                    if i in set(a[0] for a in self.nt_al):
                        return "(%s:%s)" % (label, i)
                    else:
                        return "%s:%s" % (label, i)

        return "%s -> %s" % (format_tree(self.lhs, self.lhs.roots[0]),
            " ; ".join(format_tree(self.rhs, r, True) for r in self.rhs.roots))


def extract_rules(s, t, a):

    def annotate_spans(tree):
        l = tree.length()
        tree.leaves = []
        tree.span = [set() for _ in tree]
        for index, (label, children) in enumerate(tree):
            if children:
                for ch in children:
                    tree.span[index].update(tree.span[ch])
            else:
                tree.span[index].add(index) # l - len(self.leaves) - 1
                tree.leaves.insert(0, index)

    def follow_targets(targets, span):
        new_targets = set()
        for target in targets:
            while True:
                parent = t.parent(target)
                if parent is None or (ts_span[parent] - span):
                    new_targets.add(target)
                    break
                else:
                    target = parent
        return new_targets

    annotate_spans(s)
    annotate_spans(t)

    st = defaultdict(set) # correspondence source-target
    ts = defaultdict(set) # correspondence target-source

    # Map lexical alignments from yield indices to tree nodes
    for (l1, l2) in a:
        st[s.leaves[l1]].add(t.leaves[l2])
        ts[t.leaves[l2]].add(s.leaves[l1])

    # Compute the span of each node going to the other tree
    st_span = [set().union(*[st[leaf] for leaf in s.span[index]])
        for index, node in enumerate(s)]
    ts_span = [set().union(*[ts[leaf] for leaf in t.span[index]])
        for index, node in enumerate(t)]

    # Don't extract unaligned words or words with n-1 alignment
    # This means the source-target-source span should not be empty
    # or larger than the source span
    sts_span = {}
    for index, node in enumerate(s):
        sts_span[index] = set().union(*(ts_span[i] for i in st_span[index]))
    
    # Determine cut points: every node in the source tree is
    # a cut point unless the parent node has the same span
    current_span = None
    cut_points = []

    # ALTERNATIVE: use s.span[index] to extract slightly more rules
    # (The difference is in the handling of unaligned source words)
    for index, node in enumerate(s):
        if sts_span[index]:
            if sts_span[index] == current_span:
                cut_points.pop()
            current_span = sts_span[index]
            cut_points.append(index)

    # Establish rules as (source, targets) pairs
    cut_pairs = dict((index, follow_targets(st_span[index], s.span[index]))
        for index in cut_points
        # see above for the purpose of sts_span
        if sts_span[index] and (s.span[index] >= sts_span[index]))

    # Find the (nonterminal) leaves for the rules
    rules = []
    stack = []
    for (index, node) in enumerate(s):
        label, children = node
        if children:
            rch = [stack.pop() for ch in children]
            rch = [c for ch in rch for c in ch]
            if index in cut_pairs:
                rules.append(((index, cut_pairs[index]), rch))
                stack.append([(index, cut_pairs[index])])
            else:
                stack.append(rch)
        if not children:
            stack.append([(index, st[index])])

    return [MBOTRule.from_cut_points(s, t, rule) for rule in rules]


if __name__ == "__main__":
    source_file = open(sys.argv[1])
    target_file = open(sys.argv[2])
    align_file  = open(sys.argv[3])
    triplets = izip(source_file, target_file, align_file)

    for i, (s_string, t_string, a_string) in enumerate(triplets):
        if (i + 1) % 1000 == 0:
            sys.stderr.write("\r%s" % (i + 1))
        # if forest: TODO

        if s_string.startswith("(())") or t_string.startswith("No"):
            sys.stderr.write("No parse tree in line %s\n" % (i,))
        else:
            s = TreeSequence(s_string)
            t = TreeSequence(t_string)
            a = Alignment(a_string)

            # if length doesn't match
            if a.compatible(s, t):
                rules = extract_rules(s, t, a)
                for rule in rules:
                    print rule
            else:
                sys.stderr.write("Alignment broken in line %s\n" % (i,))
        print

    source_file.close()
    target_file.close()
    align_file.close()
